@@ -2,12 +2,13 @@ import time
 from json.decoder import JSONDecodeError
 from logging import Logger
 from os import environ
-from typing import Optional, Sequence, Dict, Union, cast, Iterator, Iterable, List, Literal, Any, Tuple
+from typing import Optional, Sequence, Dict, Union, cast, Iterator, Iterable, List, Literal, Any, Tuple, Set
 
 import requests
 from api_session import APISession, escape_path
 from requests.exceptions import HTTPError
 
+from .attributes import CATALOG_PRODUCT_ENTITY_TYPE_ID
 from .exceptions import MagentoException, MagentoAssertionError
 from .queries import Query, make_search_query, make_field_value_query
 from .types import Product, SourceItem, Sku, Category, MediaEntry, MagentoEntity, Order, PathId, Customer, SourceItemIn, \
@@ -24,7 +25,7 @@ DEFAULT_ATTRIBUTE_DICT = {
     "apply_to": [],
     "backend_type": "int",
     "custom_attributes": [],
-    "entity_type_id": "4",
+    "entity_type_id": str(CATALOG_PRODUCT_ENTITY_TYPE_ID),
     "extension_attributes": {},
     "frontend_input": "select",
     "is_comparable": False,
@@ -388,9 +389,10 @@ class Magento(APISession):
                                         **kwargs)
         return ok
 
-    def async_add_products_to_categories(self, category_product_links: Iterable[Dict[str, Any]], **kwargs: Any) -> Any:
+    def async_add_products_to_categories(self, category_product_links: Iterable[Dict[str, Any]], **kwargs: Any) \
+            -> MagentoEntity:
         """Asynchronously add products to categories."""
-        return self.post_json_api(
+        res: MagentoEntity = self.post_json_api(
             "/V1/categories/byCategoryId/products",
             json=[{"productLink": category_product_link}
                   for category_product_link in category_product_links],
@@ -398,13 +400,15 @@ class Magento(APISession):
             throw=True,
             **kwargs,
         )
+        return res
 
-    def async_remove_products_from_categories(self, category_ids_skus: Iterable[Tuple[int, str]], **kwargs: Any) -> Any:
+    def async_remove_products_from_categories(self, category_ids_skus: Iterable[Tuple[int, str]], **kwargs: Any) \
+            -> MagentoEntity:
         """Asynchronously remove products from categories.
 
         :param category_ids_skus: Pairs of (category_id, sku).
         """
-        return self.delete_json_api(
+        res: MagentoEntity = self.delete_json_api(
             "/V1/categories/byCategoryId/products/bySku",
             json=[{"categoryId": category_id, "sku": sku}
                   for category_id, sku in category_ids_skus],
@@ -412,6 +416,7 @@ class Magento(APISession):
             throw=True,
             **kwargs,
         )
+        return res
 
     # CMS
     # ===
@@ -563,24 +568,25 @@ class Magento(APISession):
     # ========
 
     def create_order_invoice(self, order_id: PathId, payload: Optional[Dict[str, Any]] = None, notify: bool = True,
-                             **kwargs: Any) -> Any:
+                             **kwargs: Any) -> str:
         """Create an invoice for an order.
 
         See:
-        * https://devdocs.magento.com/guides/v2.4/rest/tutorials/orders/order-create-invoice.html
-        * https://www.rakeshjesadiya.com/create-invoice-using-rest-api-magento-2/
+          * https://developer.adobe.com/commerce/webapi/rest/tutorials/orders/order-create-invoice/
+          * https://www.rakeshjesadiya.com/create-invoice-using-rest-api-magento-2/
 
         :param order_id: Order id.
         :param payload: payload to send to the API.
-        :param notify: if True (default), notify the client. This is overridden by ``payload``.
-        :return:
+        :param notify: if True (default), notify the client. This is equivalent to ``payload={"notify": True}``.
+        :return: invoice ID, as a number in a string
         """
         if payload is None:
             payload = {}
 
         payload.setdefault("notify", notify)
 
-        return self.post_json_api(f"/V1/order/{escape_path(order_id)}/invoice", json=payload, **kwargs)
+        invoice_id: str = self.post_json_api(f"/V1/order/{escape_path(order_id)}/invoice", json=payload, **kwargs)
+        return invoice_id
 
     def get_invoice(self, invoice_id: int, *,
                     none_on_404: bool = False,
@@ -701,12 +707,13 @@ class Magento(APISession):
         ok: bool = self.post_json_api(f"/V1/orders/{escape_path(order_id)}/unhold", **kwargs)
         return ok
 
-    def save_order(self, order: Order, **kwargs: Any) -> Any:
+    def save_order(self, order: Order, **kwargs: Any) -> MagentoEntity:
         """Save an order."""
-        return self.post_api("/V1/orders", json={"entity": order}, **kwargs)
+        saved_order: MagentoEntity = self.post_json_api("/V1/orders", json={"entity": order}, **kwargs)
+        return saved_order
 
     def set_order_status(self, order: Order, status: str, *, external_order_id: Optional[str] = None,
-                         **kwargs: Any) -> Any:
+                         **kwargs: Any) -> MagentoEntity:
         """Change the status of an order, and optionally set its ``ext_order_id``. This is a convenient wrapper around
         ``save_order``.
         Note it does not check if orders are on hold before, and may result in invalid states where an order has a state
@@ -777,12 +784,14 @@ class Magento(APISession):
     # --------------
 
     def get_special_prices(self, skus: Sequence[Sku], *, store_id: Union[int, None] = None,
+                           deduplicate: bool = True,
                            **kwargs: Any) -> List[MagentoEntity]:
         """Get special prices for a sequence of SKUs.
 
         :param skus:
         :param store_id: Filter by store ID.
           This is done on the response as Magento doesn’t support this filter in the REST API.
+        :param deduplicate: if True (default), remove duplicated prices.
         :param kwargs:
         :return:
         """
@@ -792,6 +801,22 @@ class Magento(APISession):
             special_prices = [special_price for special_price in special_prices
                               if special_price["store_id"] == store_id]
 
+        if deduplicate:
+            seen: Set[Tuple[Tuple[str, Any], ...]] = set()
+
+            # For some reason special_prices can be duplicated,
+            # especially when having multiple stores with a special_price in each one
+            deduplicated_special_prices: List[MagentoEntity] = []
+
+            for special_price in special_prices:
+                ks = tuple(sorted(special_price.items()))
+                if ks not in seen:
+                    deduplicated_special_prices.append(special_price)
+
+                seen.add(ks)
+
+            special_prices = deduplicated_special_prices
+
         return special_prices
 
     def save_special_prices(self, special_prices: Sequence[MagentoEntity], **kwargs: Any) \
@@ -989,19 +1014,21 @@ class Magento(APISession):
         # https://magento.redoc.ly/2.3.6-admin/tag/productssku#operation/catalogProductRepositoryV1DeleteByIdDelete
         return cast(bool, response.json())
 
-    def async_update_products(self, product_updates: Iterable[MagentoEntity], **kwargs: Any) -> Any:
+    def async_update_products(self, product_updates: Iterable[MagentoEntity], **kwargs: Any) -> MagentoEntity:
         """Update multiple products using the async bulk API.
 
         Example:
             >>> Magento().async_update_products([{"sku": "SK123", "name": "Abc"}, {"sku": "SK4", "name": "Def"}])
 
-        See https://devdocs.magento.com/guides/v2.4/rest/bulk-endpoints.html
+        See https://developer.adobe.com/commerce/webapi/rest/use-rest/bulk-endpoints/
 
         :param product_updates: sequence of product data dicts. They MUST contain an `sku` key.
         :return:
         """
         payload = [{"product": product_update} for product_update in product_updates]
-        return self.put_json_api("/V1/products/bySku", json=payload, async_bulk=True, **kwargs)
+        # https://developer.adobe.com/commerce/webapi/rest/use-rest/bulk-endpoints/#responses
+        res: MagentoEntity = self.put_json_api("/V1/products/bySku", json=payload, async_bulk=True, **kwargs)
+        return res
 
     def get_product_stock_item(self, sku: Sku, *, none_on_404: bool = False, none_on_empty: bool = False,
                                **kwargs: Any) -> MagentoEntity:
@@ -1055,25 +1082,26 @@ class Magento(APISession):
                                                **kwargs)
         return ret
 
-    def link_child_product(self, parent_sku: Sku, child_sku: Sku, **kwargs: Any) -> Any:
+    def link_child_product(self, parent_sku: Sku, child_sku: Sku, **kwargs: Any) -> bool:
         """Link two products, one as the parent of the other.
 
         :param parent_sku: SKU of the parent product
         :param child_sku: SKU of the child product
-        :return: `requests.Response` object
         """
-        return self.post_json_api(f"/V1/configurable-products/{escape_path(parent_sku)}/child",
-                                  json={"childSku": child_sku}, **kwargs)
+        ok: bool = self.post_json_api(f"/V1/configurable-products/{escape_path(parent_sku)}/child",
+                                      json={"childSku": child_sku}, **kwargs)
+        return ok
 
-    def unlink_child_product(self, parent_sku: Sku, child_sku: Sku, **kwargs: Any) -> Any:
+    def unlink_child_product(self, parent_sku: Sku, child_sku: Sku, **kwargs: Any) -> bool:
         """Opposite of link_child_product().
 
         :param parent_sku: SKU of the parent product
         :param child_sku: SKU of the child product
         """
-        return self.delete_json_api(
+        ok: bool = self.delete_json_api(
             f"/V1/configurable-products/{escape_path(parent_sku)}/children/{escape_path(child_sku)}",
             **kwargs)
+        return ok
 
     def save_configurable_product_option(self, sku: Sku, option: MagentoEntity, **kwargs: Any) -> int:
         """Save a configurable product option.
@@ -1166,6 +1194,27 @@ class Magento(APISession):
     # Shipments
     # =========
 
+    def get_shipment(self, shipment_id: PathId, **kwargs: Any) -> Union[MagentoEntity, None]:
+        """Get a shipment."""
+        shipment: Union[MagentoEntity, None] = self.get_json_api(f"/V1/shipment/{escape_path(shipment_id)}",
+                                                                 none_on_404=True,
+                                                                 **kwargs)
+        return shipment
+
+    def get_shipment_label(self, shipment_id: PathId, **kwargs: Any) -> str:
+        """Get a shipment label.
+
+        If the shipment doesn't exist or it doesn't have a label, the returned label is an empty string.
+        """
+        label: str = self.get_json_api(f"/V1/shipment/{escape_path(shipment_id)}/label", **kwargs)
+        return label
+
+    def get_shipment_comments(self, shipment_id: PathId, **kwargs: Any) -> Iterator[MagentoEntity]:
+        """Get the comments for a shipment."""
+        comments: Iterator[MagentoEntity] = self.get_paginated(f"/V1/shipment/{escape_path(shipment_id)}/comments",
+                                                               **kwargs)
+        return comments
+
     def get_shipments(self, *, query: Query = None, limit: int = -1, **kwargs: Any) -> Iterator[MagentoEntity]:
         """Return shipments (generator)."""
         shipments: Iterator[MagentoEntity] = self.get_paginated("/V1/shipments", query=query, limit=limit, **kwargs)
@@ -1175,6 +1224,7 @@ class Magento(APISession):
         """Ship an order.
 
         Return the shipment ID as a string.
+        Throws a `MagentoException` in case of error.
         """
         response = self.post_api(f"/V1/order/{order_id}/ship", json=payload, throw=True,
                                  **kwargs)
@@ -1191,7 +1241,7 @@ class Magento(APISession):
 
         return body
 
-    def get_order_shipments(self, order_id: Union[int, str], **kwargs: Any) -> Any:
+    def get_order_shipments(self, order_id: Union[int, str], **kwargs: Any) -> Iterator[MagentoEntity]:
         """Get shipments for the given order id."""
         return self.get_shipments(query=make_field_value_query("order_id", order_id), **kwargs)
 
@@ -1300,8 +1350,9 @@ class Magento(APISession):
     def save_source(self, source: MagentoEntity, **kwargs: Any) -> Any:
         """Save a source.
 
-        https://adobe-commerce.redoc.ly/2.4.6-admin/tag/inventorysources/#operation/PostV1InventorySources
+        https://adobe-commerce.redoc.ly/2.4.8-admin/tag/inventorysources#operation/PostV1InventorySources
         """
+        # The docs don't specify the return type
         return self.post_json_api("/V1/inventory/sources", json={"source": source}, **kwargs)
 
     # Source Items
@@ -1337,13 +1388,10 @@ class Magento(APISession):
                     self.get_paginated("/V1/inventory/source-items", query=query, limit=limit, **kwargs))
 
     def save_source_items(self, source_items: Sequence[Union[SourceItem, SourceItemIn]], **kwargs: Any) -> Any:
-        """Save a sequence of source-items. Return None if the sequence is empty.
-
-        :param source_items:
-        :return:
-        """
+        """Save a sequence of source-items. Return None if the sequence is empty."""
         if not source_items:
             return None
+        # The docs don't specify the return type
         return self.post_json_api("/V1/inventory/source-items", json={"sourceItems": source_items}, **kwargs)
 
     def delete_source_items(self, source_items: Iterable[Union[SourceItem, SourceItemIn]], **kwargs: Any) -> Any:
@@ -1359,7 +1407,7 @@ class Magento(APISession):
         payload = {
             "sourceItems": [{"sku": s["sku"], "source_code": s["source_code"]} for s in source_items],
         }
-        # Note the docs are not saying what's the normal return value
+        # The docs don't specify the return type
         return self.post_json_api("/V1/inventory/source-items-delete", json=payload, **kwargs)
 
     def delete_source_items_by_source_code(self, source_code: str, **kwargs: Any) -> Any:
