@@ -1,6 +1,4 @@
-from typing import Any, Dict
-
-from magento.types import Order
+from magento.types import Order, ShippingAddress
 
 __all__ = (
     'is_order_on_hold',
@@ -9,21 +7,24 @@ __all__ = (
 )
 
 
-def is_order_on_hold(order: Order):
+def is_order_on_hold(order: Order) -> bool:
     """Test if an order is on hold."""
     return order["status"] == "holded" or "hold_before_state" in order
 
 
-def is_order_cash_on_delivery(order: Order):
+def is_order_cash_on_delivery(order: Order) -> bool:
     """Test if an order is paid with 'cash-on-delivery'."""
     # From Magento\OfflinePayments\Model\Cashondelivery::PAYMENT_METHOD_CASHONDELIVERY_CODE
-    return order["payment"]["method"] == 'cashondelivery'
+    # For some reason Mypy is too dumb to see this is a boolean
+    b: bool = order["payment"]["method"] == 'cashondelivery'
+    return b
 
 
-def get_order_shipping_address(order: Order) -> Dict[str, Any]:
-    """
-    Return the first shipping address of an order.
+def get_order_shipping_address(order: Order) -> ShippingAddress:
+    """Return the first shipping address of an order.
+
     Note the returned dict is a reference, so if you modify it, it modifies the order.
     Make a copy if you want to modify the address without affecting the order.
     """
-    return order["extension_attributes"]["shipping_assignments"][0]["shipping"]["address"]
+    address: ShippingAddress = order["extension_attributes"]["shipping_assignments"][0]["shipping"]["address"]
+    return address
